from collections import deque

from scrapgo.core import SoupParser, RequestsBase, CachedRequests
from scrapgo.core.http import RequestsBase
from scrapgo.lib import is_many_type, pluralize

from .actions import resolve_link, ReduceMixin, UrlPatternAction
from .meta import ResponseMeta



class CrawlMixin(ReduceMixin, SoupParser):
    urlorders = None

    def crawl(self, _response=None, _urlorders=None, _results=None, _responsemap=None, _composed:dict=None):
        
        action, *rest = _urlorders or self.urlorders

        _composed = _composed or dict()
        response_queue = deque([_response])

        if _urlorders == []:
            response_queue = deque()
        
        visited = set()
        while response_queue:
            response = response_queue.pop()
            
            cookies = self.dispatch_cookies(action)
            headers = self.dispatch_headers(action)
            headers.update(self.dispatch_referer(action, response))

            is_parsable = True
            
            for link in self.dispatch_renderer(action, response, _responsemap, _results):
                                
                url = resolve_link(action, link, response)
                url = self.dispatch_fields(action, url)

                if not self.dispatch_urlfilter(action, url, _responsemap, _results):
                    continue
                    
                if isinstance(action, UrlPatternAction):
                    if url in visited:
                        continue
                    visited.add(url)

                #check post method
                for payloads in self.dispatch_payloader(action, _responsemap, _results):
                    requests_kwargs = dict(url=url, headers=headers, cookies=cookies)

                    # if payloads is {}, explicitly to None
                    payloads = payloads or None

                    json_type = False
                    if content_type := headers.get('Content-Type'):
                        if 'application/json' in content_type:
                            json_type = isinstance(payloads, (dict, list))

                    requests_kwargs['json' if json_type else 'data'] = payloads

                    sub_response = self.dispatch_response(action, **requests_kwargs)

                    ## content type check
                    # soup로로 처리 불가능 한것은 content 그대로 넘김
                    soup = sub_response.content
                    is_parsable = self._is_parsable(sub_response)
                    if is_parsable:
                        soup = self._load_soup(soup)

                    ## respone meta setting
                    meta = ResponseMeta(soup=soup)
                    meta.set_urlutils(link, action)
                    meta.set_responsemap(sub_response, action)

                    if response:
                        meta.update_responsemap(response.crawler.responsemap)
                    setattr(sub_response, 'crawler', meta)

                    if _responsemap:
                        meta.update_responsemap(_responsemap)

                    ## parsing
                    extracted = self.dispatch_extractor(action, meta, sub_response, _results, _composed)
                    results = self.dispatch_parser(action, sub_response, extracted, meta, _results, _composed)

                    if is_many_type(results):
                        results = list(results)                    
                    
                    ## results proccessing
                    if hasattr(self, 'compose'):
                        _composed.setdefault(action.name ,[])
                        _composed[action.name] += pluralize(results)

                    if is_parsable is True:
                        if isinstance(action, UrlPatternAction):
                            if action.recursive:
                                response_queue.append(sub_response)
                        self.crawl(sub_response, rest, results, meta.responsemap, _composed)
                        
            # BFO if not passable
            if is_parsable is False:
                self.crawl(response, rest, results or _results, meta.responsemap, _composed)
        
        if _response is None:
            self.dispatch_compose(_composed)
    


class RequestsCrawler(CrawlMixin, RequestsBase):
    pass


class CachedRequestsCrawler(CrawlMixin, CachedRequests):
    pass



from collections import abc
from bs4 import BeautifulSoup, element

from scrapgo.lib import strcompile, prettier, is_many_type
from scrapgo.settings import FromSettings

from .exceptions import *



class SoupParser(FromSettings):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _load_soup(self, content):
        soup = self.apply_settings(BeautifulSoup, content, prefix='BS4_')
        return soup

    def _is_parsable(self, response):
        content_type_info = response.headers.get('Content-Type')
        if not content_type_info:
            return False

        for ctype in self.settings['PARSE_CONTENT_TYPES']:
            if ctype in content_type_info:
                return True
        return False
    
    def parse_linkpattern(self, content, urlpattern, remove_duplicates=True, attrs=None, css_selector=None, **kwargs):
        soup = self._load_soup(content)
        if css_selector:
            soup = soup.select_one(css_selector) or soup

        compiled = strcompile(urlpattern)

        parsed = [
            sp[attr]
            for attr in attrs or self.settings['CRAWL_TARGET_ATTRS']
            for sp in soup(attrs={attr: compiled})
        ]

        if remove_duplicates:
            parsed = list(dict.fromkeys(parsed))
        return parsed
    
    def validate_extracted(self, extracted, func, meta, _root=True):
        
        if not isinstance(meta.soup, BeautifulSoup):
            raise CannotExtractError(f"{func.__name__} won't recive soup instance or None")
 
        if isinstance(extracted, element.Tag):
            if self.settings['EXTRACT_AUTO_SOUP2TEXT']:
                extracted = extracted.get_text(strip=self.settings['EXTRACT_AUTO_STRIP'])
            if self.settings['EXTRACT_AUTO_PRETTIFY']:
                extracted = prettier(extracted)
        elif isinstance(extracted, str):
            if self.settings['EXTRACT_AUTO_STRIP']:
                extracted = extracted.strip()
            if self.settings['EXTRACT_AUTO_PRETTIFY']:
                extracted = prettier(extracted)
        elif isinstance(extracted, abc.Mapping):
            ext = extracted.__class__()
            for key, val in extracted.items():
                val = self.validate_extracted(val, func, meta, _root=False)
                if self.settings['EXTRACT_AUTO_PRETTIFY']:
                    key = prettier(key)
                ext[key] = val
            return ext
        elif is_many_type(extracted):
            extracted_list = [
                self.validate_extracted(ext, func, meta, _root=False) for ext in extracted
            ]
            return extracted_list
        return extracted



import functools, time

from requests.exceptions import ProxyError, ConnectTimeout
from fake_useragent import FakeUserAgent

from scrapgo.lib import format_bytes_size
from .exceptions import *



def set_user_agent(header, agent_name, **fkagent_kwargs):
    fagent = FakeUserAgent(**fkagent_kwargs)
    if 'user-agent' not in map(str.lower, header):
        header['User-Agent'] = getattr(fagent, agent_name)


def as_reqeusts_kwargs(**kwargs):
    params = ['method', 'url', 'params', 'data', 'headers', 'cookies', 'files', 'auth', 'timeout', 'allow_redirects', 'proxies', 'hooks', 'stream', 'verify', 'cert', 'json']
    return {
        k:v for k, v in kwargs.items()
        if k in params
    }


def _get_pre_request_log(method, url, data=None, proxies=None, **extra):
    METHOD = method.upper()
    PROXIES = f"proxy:{proxies}" if proxies else '' 
    prefix = ' '.join(filter(None, [METHOD, url, PROXIES]))
    PAYLOADS = f"payloads:{format_bytes_size(payloads, 2)}" if (data:=data) and (payloads := len(data)) else ''
    
    if postfix:= list(filter(None, [PAYLOADS])):
        postfix = ', '.join(postfix)
        postfix = f' ({postfix})'
    else:
        postfix = ''

    log = f"{prefix}{postfix}"
    return log

def _get_after_request_log(response, delay):
    content_length = len(response.content or '')
    status = response.status_code
    reason = response.reason
    size_exp = format_bytes_size(content_length)
    elapsed = round(response.elapsed.microseconds / (1000 * 1000),2)

    log = f'{status} {reason} {size_exp} {elapsed}s'

    if ctype:= response.headers.get('Content-Type'):
        log = f'{log} {ctype}'

    if hasattr(response, 'from_cache'):
        if response.from_cache:
            log = f'{log} (From Cache)'
            return log
    
    log = f'{log} (delay:{delay or 0}s)'
    return log


def trace(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        delay = self.get_delay(kwargs.get('delay'))
        try:
            proxy = self.get_first_proxy()
            pre_log = _get_pre_request_log(proxies=proxy, **kwargs)
            if self.settings['REQUEST_LOGGING']:
                print(f'{pre_log}')
            r = func(self, *args, **kwargs)
        except Exception as e:
            print(e)
            raise
        else:
            log = f'  => {_get_after_request_log(r, delay)}'
            if self.settings['REQUEST_LOGGING']:
                print(log)
            if hasattr(r, 'from_cache'):
                if r.from_cache:
                    return r
            time.sleep(delay)
            return r
    return wrapper



def retry_for_raise(fetch):
    @functools.wraps(fetch)
    def wrapper(self, *args, **kwargs):
        url = kwargs.get('url')
        retry_intervals = list(self.settings['RETRY_INTERVAL_SECONDS'])
        while retry_intervals:
            try:
                response = fetch(self, *args, **kwargs)
            except Exception as e:
                if self.proxies and isinstance(e, RotateProxiesDone):
                    raise
                interval = retry_intervals.pop(0)
                print(f" - {e}")
                print(f" - Request {url} Failed, retry after {interval} sec...(retry remains: {len(retry_intervals)})")
                time.sleep(interval)
            else:
                return response
        raise RetryMaxCountDone(f"Request for {url} has failed!")
    return wrapper


def rotate_proxy(fetch):
    @functools.wraps(fetch)
    def wrapper(self, *args, **kwargs):
        proxy = self.get_first_proxy()
        if not proxy or kwargs.get('proxies'):
            return fetch(self, *args, **kwargs)

        kwargs.update({'proxies': proxy})

        rotate_count = 0
        while True:
            try:
                response = fetch(self, *args, **kwargs)
            except (ProxyError, ConnectTimeout) as e:
                next_proxy = self.rotate_proxies()
                rotate_count += 1

                if proxy == next_proxy:
                    raise RotateProxiesDone("All requests through {len(self.proxies)} proxies fail")
    
                kwargs.update({'proxies': next_proxy})
                print(f" - The requests by pass proxy server {proxy} has failed. Trying to next proxy server {next_proxy}...(rotate rate: {rotate_count}/{len(self.proxies)})")
            else:
                return response
    return wrapper



import requests_cache

from .base import RequestsBase



class CachedRequests(RequestsBase):
    cached_session_params = ['key_fn', 'ignored_parameters', 'match_headers', 'stale_if_error', 'filter_fn']

    def load_session(self):
        return self.apply_settings(
            requests_cache.CachedSession,
            prefix='REQUEST_CACHE_',
            allowed_params=self.cached_session_params
        )
    
    def delete_cache(self, **requests_kwargs):
        key = self.session.cache.create_key(**requests_kwargs)
        self.session.cache.delete(key)

    def fetch(self, *args, **kwargs):
        requests_kwargs = {
            k:v for k, v in kwargs.items()
            if k in self.request_params
        }
        if kwargs.get('refresh') is True:
            self.delete_cache(**requests_kwargs)
        return super().fetch(*args, **kwargs)





from ..lib.module import reduce_module2dict, select_kwargs
from . import base



class FromSettings:

    def __init__(self, settings=None):
        self.settings = reduce_module2dict(base)
        if settings:
            self.settings.update(reduce_module2dict(settings))
            self.settings = self.valid_settings(self.settings)

    def valid_settings(self, settings):
        self._validate_request_delay(settings)
        self._validate_retry_interval_seconds(settings)
        return settings

    def apply_settings(self, callable, *args, prefix:str=None, lowerkey:bool=True, allowed_params:list=None, **kwargs):
        if prefix:
            settings_kwargs = {
                k.replace(prefix, ''): v for k, v in self.settings.items()
                if k.startswith(prefix)
            }
        else:
            settings_kwargs = dict(self.settings)
            
        if lowerkey:
            settings_kwargs = {
                k.lower(): v for k,v in settings_kwargs.items()
            }
        updated_params = {**settings_kwargs, **kwargs}
        return select_kwargs(callable, *args, allowed_params=allowed_params, **updated_params)


    def _validate_retry_interval_seconds(self, settings):
        value = settings['RETRY_INTERVAL_SECONDS']
        if value is None:
            value = ()

        if isinstance(value, (int, float)):
            value = value, 
        
        if not isinstance(value, (tuple, list)):
            raise ValueError('RETRY_INTERVAL_SECONDS must be tuple of numbers or a number')
        settings['RETRY_INTERVAL_SECONDS'] = value


    def _validate_request_delay(self, settings):
        value = settings['REQUEST_DELAY']
        if not value:
            value = 0, 0
        elif isinstance(value, (int, float)):
            value = value, value,
        elif isinstance(value, (list, tuple)):
            if len(value) == 1:
                value = value[0], value[0]
            else:
                value = value[0], value[-1]
        else:
            raise ValueError('REQUEST_DELAY must be tuple of numbers or a number')
        settings['REQUEST_DELAY'] = value





import functools

from collections import abc



def is_many_type(value) -> bool:
    if isinstance(value, abc.Iterable):
        if isinstance(value, abc.Mapping):
            return False
        elif isinstance(value, abc.Sequence):
            if isinstance(value, (str, bytes)):
                return False
        return True
    return False
    


def singulize(value) -> abc.Generator:
    if not is_many_type(value):
        if value is not None:
            yield value
    else:
        for row in value:
            yield from singulize(row)


def pluralize(value) -> list:
    pluralized = []
    if not is_many_type(value):
        if value is not None:
            pluralized.append(value)
    else:
        for row in value:
            pluralized += pluralize(row)
    return pluralized


def assingles(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        r = func(*args, **kwargs)
        yield from singulize(r)
    return wrapper




from glob import glob
from dataclasses import dataclass



@dataclass
class BaseAction:
    name:str
    parser:... = None
    extractor:... = None
    refresh:bool = False
    delay:int = None

    def __post_init__(self, *args, **kwargs):
        if not self.name:
            raise NotImplementedError("action name must be specified")

    def as_kwargs(self):
        return self.__dict__


@dataclass
class HttpAction(BaseAction):
    headers:dict=None
    cookies:dict=None
    referer:str=None
    payloader:...=None
    urlfilter:...=None
    fields:list=None
    method:str=None



@dataclass
class UrlRenderAction(HttpAction):
    url:str = None
    urlrenderer:... = None

    def __post_init__(self, *args, **kwargs):
        if not (self.url or self.urlrenderer):
            raise NotImplementedError("an url or urlrenderer must be specified")
    


@dataclass
class UrlPatternAction(HttpAction):
    urlpattern:str=None
    urlpattern_renderer:...=None
    remove_duplicates:bool=None
    attrs:dict=None
    css_selector:str=None
    recursive:bool=False

    def __post_init__(self, *args, **kwargs):
        if not (self.urlpattern or self.urlpattern_renderer):
            raise NotImplementedError("urlpattern or urlpattern_renderer must be specified")


@dataclass
class FileAction(BaseAction):
    path:str=None
    path_renderer:...=None
    method:str='get'

    def __post_init__(self):
        if not (self.path or self.path_renderer):
            raise NotImplementedError("path or path_renderer must be specified")
        self._check_globpath()

    def _check_globpath(self):
        if '*' in self.path:
            self.path_renderer = self.path_renderer or (lambda: glob(self.path, recursive=True))



@dataclass
class CurlAction(BaseAction):
    command:str=None
    command_renderer:...=None

    def __post_init__(self):
        if not (self.command or self.command_renderer):
            raise NotImplementedError("command or command_renderer must be specified")



def urlrender(
    url:str=None, 
    urlrenderer:...=None,
    fields:list=None,
    payloader:...=None,
    parser:...=None,
    extractor:...=None,
    urlfilter:...=None,
    headers:dict=None, referer:str=None, cookies:dict=None,
    method:str=None,
    refresh:bool=False, delay:int=None,
    name:str=None):

    return UrlRenderAction(**locals())

def urlpattern(
    urlpattern:str=None,
    urlpattern_renderer:...=None,
    fields:list=None,
    payloader:...=None,
    parser:...=None,
    extractor:...=None,
    urlfilter:...=None,
    remove_duplicates:bool=None,
    attrs:dict=None,
    css_selector:str=None,
    recursive:bool=False,
    headers:dict=None, referer:str=None, cookies:dict=None, method:str=None,
    refresh:bool=False, delay:int=None,
    name:str=None):
    return UrlPatternAction(**locals())


def curl(
    command:str=None,
    command_renderer:...=None,
    parser:...=None, extractor:...=None,
    refresh:bool=False, delay:int=None,
    name:str=None):
    return CurlAction(**locals())


def file(
    path:str=None,
    path_renderer:...=None,
    parser:...=None,
    extractor:...=None,
    refresh:bool=False,
    delay:int=None,
    name:str=None):
    return FileAction(**locals())



class RequestErrorBase(Exception):
    '''Request Exception Base
    '''


class RetryMaxCountDone(RequestErrorBase):
    '''Retry maximum count over
    '''


class ResponseHasNoContent(RequestErrorBase):
    '''Response has no contents
    '''


class RotateProxiesDone(RequestErrorBase):
    '''Rotate all Proxies list
    '''


from collections import abc
from scrapgo.lib import parse_query, extgroup
from dataclasses import dataclass, field

from .actions import UrlPatternAction, UrlRenderAction


@dataclass
class ResponseMeta:
    match:callable = None
    pattern:str = None
    query:dict = None
    soup:... = None
    responsemap:dict = field(default_factory=dict)


    def set_urlutils(self, link, action):
        if isinstance(action, UrlRenderAction):
            if isinstance(link, str):
                self.query = parse_query(link)
            elif isinstance(link, abc.Mapping):
                self.query = link
            else:
                self.query = None
        
        elif isinstance(action, UrlPatternAction):
            self.match = extgroup(action.urlpattern, link)
            self.query = parse_query(link)
    
    def set_responsemap(self, response, action):
        self.responsemap = {
            action.name: response
        }
    
    def update_responsemap(self, prev_response_meta):
        self.responsemap.update(
            prev_response_meta
        )


import requests
from requests.structures import CaseInsensitiveDict

from scrapgo.settings import FromSettings
from scrapgo.lib import gen_random_value, get_rotate

from .utils import trace, retry_for_raise, rotate_proxy, set_user_agent



class RequestParamsMixin:
    HEADERS = None
    COOKIES = None
    PROXIES = None

    def init_parmas(self):
        self.headers = CaseInsensitiveDict(self.HEADERS)
        self.cookies = self.COOKIES or {}
        self.proxies = self.PROXIES or []
        if hasattr(self, 'settings'):
            if proxy := self.settings.get('REQUESTS_PROXY'):
                self.proxies.append(proxy)

        self.set_user_agent()

    def set_user_agent(self):
        self.apply_settings(
            set_user_agent, self.headers, prefix='FAKE_USER_AGENT_'
        )

    def get_headers(self) -> CaseInsensitiveDict:
        return CaseInsensitiveDict(self.headers)

    def set_header(self, headers:dict):
        self.headers = CaseInsensitiveDict(headers)
    
    def get_cookies(self):
        return dict(self.cookies)
    
    def set_cookies(self, cookies):
        self.cookies = cookies

    def get_first_proxy(self):
        for proxy in self.proxies:
            return proxy

    def rotate_proxies(self):
        if self.proxies:
            self.proxies = get_rotate(self.proxies)
            return self.proxies[0]



class RequestsBase(RequestParamsMixin, FromSettings):
    request_params = ['method', 'url', 'params', 'data', 'headers', 'cookies', 'files', 'auth', 'timeout', 'allow_redirects', 'proxies', 'hooks', 'stream', 'verify', 'cert', 'json']

    def __init__(self, settings=None):
        super().__init__(settings=settings)
        self.init_parmas()
        self.session = self.load_session()
    
    def get_delay(self, delay=None):
        if delay is not None:
            return delay
        return gen_random_value(*self.settings['REQUEST_DELAY'])

    def load_session(self):
        return requests.Session()

    @trace
    @retry_for_raise
    @rotate_proxy
    def fetch(self, *args, **kwargs):

        response = self.apply_settings(
            self.session.request, prefix='REQUESTS_', 
            *args, **kwargs,
            allowed_params=self.request_params,
        )

        if response.status_code not in self.settings.get('ALLOWABLE_ERROR_STATUS_CODES', []):
            response.raise_for_status()
        return response




import shlex, re, argparse, functools
from requests.structures import CaseInsensitiveDict


def parse_curl(command):
    tokens = shlex.split(command)
    return list(filter(None, map(str.strip, tokens)))
    

def create_parser(tokens):
    parser = argparse.ArgumentParser()
    parser.add_argument('appname')
    parser.add_argument('url')
    parser.add_argument('-X', '--method',)
    parser.add_argument('-H', '--headers', action='append', default=[])
    parser.add_argument('-d', '--data', '--data-binary', '--data-raw', '--data-urlencode', '--form', default=None)
    parser.add_argument('--connect-timeout')
    parser.add_argument('-k','--insecure', action='store_true', default=True)
    parser.add_argument('--compressed', action='store_true')
    return parser


def process_args(parser, tokens):
    def _parse(pat, v):
        if g:= pat.search(v):
            return {g.group('key'): g.group('value')}
        return {}

    def _parse_headers(h):
        pat = re.compile(r'(?P<key>.+?)\:\s*(?P<value>.+)')
        return _parse(pat, h)

    def _parse_cookies(c):
        pat = re.compile(r'(?P<key>.+?)=(?P<value>.+)')
        return _parse(pat, c)

    args = parser.parse_args(tokens)
    headers = functools.reduce(lambda acc, cur: {**acc , **_parse_headers(cur)}, args.headers, CaseInsensitiveDict())

    if cookies := headers.get('cookie'):
        headers.pop('cookie')
        cookies = cookies.split(';')
        cookies=functools.reduce(lambda acc, cur: {**acc, **_parse_cookies(cur)}, cookies, {})

    return dict(
        data=args.data,
        url=args.url,
        method=args.method or ('post' if  args.data else 'get'),
        timeout=args.connect_timeout,
        headers=functools.reduce(lambda acc, cur: {**acc, **_parse_headers(cur)}, args.headers, CaseInsensitiveDict()),
        cookies=cookies,
        verify=args.insecure,
    )


def curl2requests(curl):
    tokens = parse_curl(curl)
    parser = create_parser(tokens)
    kwargs = process_args(parser, tokens)
    return kwargs


from urllib.error import URLError

from scrapgo.lib import select_kwargs, filter_params, find_function, curl2requests, assingles
from .base import UrlRenderAction, UrlPatternAction, FileAction, CurlAction
from .exceptions import *



class ReduceMixin:

    def get_action(self, name):
        for action in self.urlorders:
            if action.name == name:
                return action    

    def dispatch_response(self, action, **kwargs):
        if isinstance(action, CurlAction):
            command = kwargs.get('url')
            request_kwargs = curl2requests(command)
            return self.fetch(refresh=action.refresh, delay=action.delay, **request_kwargs)

        method = action.method or ('post' if kwargs.get('data') else 'get')
        return self.fetch(method=method, refresh=action.refresh, delay=action.delay, **kwargs)

    @assingles
    def dispatch_renderer(self, action, response, responsemap, prev_results):
        if isinstance(action, UrlRenderAction):
            if urlrenderer := action.urlrenderer:
                return self.dispatch(
                    'urlrenderer', urlrenderer,
                    url=action.url, parent_response=response, responsemap=responsemap,
                    prev=prev_results
                )                
            return action.url

        elif isinstance(action, UrlPatternAction):
            if urlpattern_renderer := action.urlpattern_renderer:
                urlpattern = self.dispatch(
                    'urlpattern_renderer', urlpattern_renderer,
                    pattern=action.urlpattern, parent_response=response, responsemap=responsemap,
                    prev=prev_results
                )
                if urlpattern is None:
                    raise UrlRendererError(f"{urlpattern_renderer} must return regex pattern of url, not {urlpattern}")
            else:
                urlpattern = action.urlpattern
            kwargs = {
                **action.as_kwargs(),
                'urlpattern': urlpattern
            }
            return self.parse_linkpattern(response.content, **kwargs)
    
        elif isinstance(action, CurlAction):
            if command_renderer:= action.command_renderer:
                return self.dispatch(
                    'curlrenderer', command_renderer,
                    command=action.command, parent_resposne=response, responsemap=responsemap,
                    prev=prev_results
                )
            return action.command
        elif isinstance(action, FileAction):
            if path_renderer := action.path_renderer:
                return self.dispatch(
                    'path_renderer', path_renderer,
                    path=action.path, parent_resposne=response, responsemap=responsemap,
                    prev=prev_results
                )
            return action.path
        else:
            raise CannotFindAction(f'{action} is not memeber of renderer action')

    @assingles
    def dispatch_payloader(self, action, responsemap, prev_results):
        if not hasattr(action, 'payloader') or not action.payloader:
            return {}
        else:
            return self.dispatch(
                'payloader', action.payloader,
                responsemap=responsemap, prev=prev_results
            )
            
    def dispatch_fields(self, action, url):
        if not hasattr(action, 'fields'):
            return url
        url = filter_params(url, action.fields)
        return url

    def dispatch_headers(self, action):
        headers = self.get_headers()
        if hasattr(action, 'headers'):
            headers.update(action.headers or {})
        return headers
    
    def dispatch_cookies(self, action):
        cookies = self.get_cookies()
        if hasattr(action, 'cookies'):
            cookies.update(action.cookies or {})
        return cookies

    def dispatch_extractor(self, action, meta, response, prev_results, results_cache):
        extractset = {}
        if module := action.extractor:
            pat = r'^extract_(?P<ext>\w+)$'
            for g, func in find_function(module, pat):
                extracted = select_kwargs(
                    func,
                    response=response,
                    meta=meta,
                    soup=meta.soup,
                    prev=prev_results,
                    results_set=results_cache
                )
                extractset[g('ext')] = self.validate_extracted(extracted, func, meta)
        return extractset

    def dispatch_parser(self, action, response, extracted, meta, prev_results, results_cache):
        results = self.dispatch(
            'parser', action.parser,
            response=response, parsed=extracted, meta=meta,
            prev=prev_results,
            results_set=results_cache
        )
        return results

    def dispatch_urlfilter(self, action, url, responsemap, prev_results):
        if not hasattr(action, 'urlfilter'):
            return True

        return self.dispatch(
            'urlfilter', action.urlfilter,
            url=url, responsemap=responsemap, prev=prev_results
        )
    
    def dispatch_referer(self, action, response):
        headers = self.get_headers()
        if not hasattr(action, 'referer'):
            return headers
        
        if action.referer and response:
            try:
                referer = response.crawler.responsemap[action.referer]
            except KeyError:
                raise CannotFindAction(f"An action named {action.referer} could not be found at urlorders.")
            else:
                headers['Referer'] = referer.url
        return headers


    def dispatch_compose(self, results_set):
        if hasattr(self, 'compose'):
            composed = self.dispatch(
                'compose', self.compose,
                **results_set
            )
            return composed


    def dispatch(self, type, func, *args, **kwargs):
        if callable(func):
            f = func
        elif isinstance(func, str):
            if hasattr(self, func):
                f = getattr(self, func)
            else:
                raise NotImplementedError(f"The method {func} is not implemented!")
        else:
            f = {
                'urlfilter': self.default_urlfilter,
                'parser': self.default_parser,
                'urlrenderer': self.default_urlrenderer,
                'urlpattern_renderer': self.default_pattern_renderer,
                'breaker': self.default_breaker,
                'payloader': self.default_payloader,
                'curlrenderer': self.default_curlrenderer,
                'path_renderer': self.default_path_renderer,
            }[type]

        return select_kwargs(f, *args, **kwargs)
    

    def default_urlfilter(self, url, **kwargs):
        return True
    
    def default_parser(self, response, **kwargs):
        return
    
    def default_urlrenderer(self, url, response, **kwargs):
        yield url
    
    def default_curlrenderer(self, curl, resposne, **kwargs):
        return curl
    
    def default_path_renderer(self, path, response, **kwargs):        
        return path
        
    def default_pattern_renderer(self, pattern, resposne, **kwargs):
        return pattern
    
    def default_breaker(self, response, **kwargs):
        return False
    
    def default_payloader(self):
        yield
    


import os
from datetime import timedelta


# Request
# USER_AGENT_NAME = 'chrome'
REQUEST_DELAY = 0.5, 1.5
RETRY_INTERVAL_SECONDS = 10, 15, 20
ALLOWABLE_ERROR_STATUS_CODES = [] 

REQUEST_LOGGING = True

# FAKE_USER_AGENT
# FAKE_USER_AGENT_USE_CACHE_SERVER = True
FAKE_USER_AGENT_AGENT_NAME = 'chrome'


# Request Cache
REQUEST_CACHE_CACHE_NAME = 'crawlite.sqlite'
REQUEST_CACHE_BACKEND = os.environ.get('REQUEST_CACHE_BACKEND', 'sqlite')
REQUEST_CACHE_EXPIRE_AFTER = timedelta(days=100)
REQUEST_CACHE_ALLOWABLE_METHODS = 'GET', 'POST',
REQUEST_CACHE_ALLOWABLE_CODES = 200,
REQUEST_CACHE_OLD_DATA_ON_ERROR = False
REQUEST_CACHE_CACHE_CONTROL = False

# requests
REQUESTS_TIMEOUT = 5
REQUESTS_PROXY = None

# SoupParser
CRAWL_TARGET_ATTRS = ['href', 'src',]
PARSE_CONTENT_TYPES = [
    'text/css', 'text/html', 'text/javascript', 'text/plain', 'text/xml',
    'application/json'
]

BS4_FEATURES = 'html.parser'
BS4_PARSE_ONLY = None
BS4_FROM_ENCODING = 'utf-8'
BS4_EXCLUDE_ENCODINGS = None
BS4_ELEMENT_CLASSES = None


EXTRACT_AUTO_STRIP = True
EXTRACT_AUTO_SOUP2TEXT = True
EXTRACT_AUTO_PRETTIFY = True

CRAWL_SUSPENDE_LOOP_POLLING_RATE = 1